@@ -1,6 +1,6 @@
 package Net::SFTP::Foreign::Backend::Unix;
 
-our $VERSION = '1.58_07';
+our $VERSION = '1.63_07';
 
 use strict;
 use warnings;
@@ -22,7 +22,7 @@ sub _defaults {
 }
 
 sub _init_transport_streams {
-    my ($self, $sftp) = @_;
+    my (undef, $sftp) = @_;
     for my $dir (qw(ssh_in ssh_out)) {
 	binmode $sftp->{$dir};
 	my $flags = fcntl($sftp->{$dir}, F_GETFL, 0);
@@ -50,6 +50,7 @@ sub _ipc_open2_bug_workaround {
 }
 
 sub _open3 {
+    my $backend = shift;
     my $sftp = shift;
     if (defined $_[2]) {
 	my $sftp_err = $_[2];
@@ -70,7 +71,7 @@ sub _open3 {
 }
 
 sub _init_transport {
-    my ($class, $sftp, $opts) = @_;
+    my ($backend, $sftp, $opts) = @_;
 
     my $transport = delete $opts->{transport};
 
@@ -98,6 +99,7 @@ sub _init_transport {
 	my $stderr_discard = delete $opts->{stderr_discard};
 	my $stderr_fh = ($stderr_discard ? undef : delete $opts->{stderr_fh});
         my $open2_cmd = delete $opts->{open2_cmd};
+        my $ssh_cmd_interface = delete $opts->{ssh_cmd_interface};
 
 	my @open2_cmd;
         if (defined $open2_cmd) {
@@ -111,11 +113,10 @@ sub _init_transport {
             $ssh_cmd = 'ssh' unless defined $ssh_cmd;
             @open2_cmd = ($ssh_cmd);
 
-            my $ssh_cmd_interface = delete $opts->{ssh_cmd_interface};
             unless (defined $ssh_cmd_interface) {
-                $ssh_cmd_interface = ( $ssh_cmd =~ /\bplink(?:\.exe)?$/i
-                                       ? 'plink'
-                                       : 'ssh');
+                $ssh_cmd_interface = ( $ssh_cmd =~ /\bplink(?:\.exe)?$/i ? 'plink'  :
+                                       $ssh_cmd =~ /\bsshg3$/i           ? 'tectia' :
+                                                                           'ssh'    );
             }
 
             my $port = delete $opts->{port};
@@ -125,6 +126,7 @@ sub _init_transport {
             my $more = delete $opts->{more};
             carp "'more' argument looks like if it should be splited first"
                 if (defined $more and !ref($more) and $more =~ /^-\w\s+\S/);
+            my @more = _ensure_list $more;
 
             if ($ssh_cmd_interface eq 'plink') {
                 $pass and !$passphrase
@@ -134,19 +136,32 @@ sub _init_transport {
             elsif ($ssh_cmd_interface eq 'ssh') {
                 push @open2_cmd, -p => $port if defined $port;
 		if ($pass and !$passphrase) {
-		    push @open2_cmd, (-o => 'NumberOfPasswordPrompts=1',
-				      -o => 'PreferredAuthentications=keyboard-interactive,password');
+		    push @open2_cmd, -o => 'NumberOfPasswordPrompts=1';
+                    push @open2_cmd, -o => 'PreferredAuthentications=keyboard-interactive,password'
+                        unless grep { $more[$_] eq '-o' and
+                                      $more[$_ + 1] =~ /^PreferredAuthentications\W/ } 0..$#more-1;
 		}
             }
+            elsif ($ssh_cmd_interface eq 'tectia') {
+            }
             else {
                 die "Unsupported ssh_cmd_interface '$ssh_cmd_interface'";
             }
             push @open2_cmd, -l => $user if defined $user;
-            push @open2_cmd, _ensure_list($more) if defined $more;
+            push @open2_cmd, @more;
             push @open2_cmd, $host;
 	    push @open2_cmd, ($ssh1 ? "/usr/lib/sftp-server" : -s => 'sftp');
         }
-	_debug "ssh cmd: @open2_cmd\n" if ($debug and $debug & 1);
+
+        my $redirect_stderr_to_tty = ( (defined $pass or defined $passphrase) and
+                                       (delete $opts->{redirect_stderr_to_tty} or
+                                        $ssh_cmd_interface eq 'tectia' ) );
+
+        $redirect_stderr_to_tty and ($stderr_discard or $stderr_fh)
+            and croak "stderr_discard or stderr_fh can not be used together with password/passphrase "
+                          . "authentication when Tectia client is used";
+
+	$debug and $debug & 1 and _debug "ssh cmd: @open2_cmd\n";
 
 	%$opts and return; # Net::SFTP::Foreign will find the
                            # unhandled options and croak
@@ -156,7 +171,7 @@ sub _init_transport {
         }
 
 	if ($stderr_discard) {
-	    $stderr_fh = $class->_open_dev_null($sftp) or return;
+	    $stderr_fh = $backend->_open_dev_null($sftp) or return;
 	}
 
         my $this_pid = $$;
@@ -185,7 +200,9 @@ sub _init_transport {
 		$expect->raw_pty(1);
 		$expect->log_user($expect_log_user);
 
-		$child = _open3($sftp, $sftp->{ssh_in}, $sftp->{ssh_out}, $stderr_fh, '-');
+                $redirect_stderr_to_tty and $stderr_fh = $pty->slave;
+
+		$child = $backend->_open3($sftp, $sftp->{ssh_in}, $sftp->{ssh_out}, $stderr_fh, '-');
 
 		if (defined $child and !$child) {
 		    $pty->make_slave_controlling_terminal;
@@ -196,6 +213,9 @@ sub _init_transport {
 		# $pty->close_slave();
 	    }
 	    else {
+                $redirect_stderr_to_tty and
+                    croak "In order to support password/passphrase authentication with the Tectia client, " .
+                        "IPC::Open3 version 1.0105 is required (current version is $IPC::Open3::VERSION)";
 		$expect = Expect->new;
 		$expect->raw_pty(1);
 		$expect->log_user($expect_log_user);
@@ -231,21 +251,26 @@ sub _init_transport {
 	    $expect->close_slave();
         }
         else {
-	    $sftp->{pid} = _open3($sftp, $sftp->{ssh_in}, $sftp->{ssh_out}, $stderr_fh, @open2_cmd);
+	    $sftp->{pid} = $backend->_open3($sftp, $sftp->{ssh_in}, $sftp->{ssh_out}, $stderr_fh, @open2_cmd);
             _ipc_open2_bug_workaround $this_pid;
 
             unless (defined $sftp->{pid}) {
                 $sftp->_conn_failed("Bad ssh command", $!);
                 return;
             }
+            # do not propagate signals sent from the terminal to the
+            # slave SSH:
+            eval {
+                setpgrp($sftp->{pid}, 0);
+            };
         }
     }
-    $class->_init_transport_streams($sftp);
+    $backend->_init_transport_streams($sftp);
 }
 
 
 sub _do_io {
-    my ($self, $sftp, $timeout) = @_;
+    my (undef, $sftp, $timeout) = @_;
 
     $debug and $debug & 32 and _debug(sprintf "_do_io connected: %s", $sftp->{_connected} || 0);
 
@@ -295,11 +320,13 @@ sub _do_io {
 			    64 * 1024, $!);
 		    $debug & 2048 and $written and _hexdump(substr($$bout, 0, $written));
 		}
-                unless ($written) {
+                if ($written) {
+                    substr($$bout, 0, $written, '');
+                }
+                elsif ($! != Errno::EAGAIN() and $! != Errno::EINTR()) {
                     $sftp->_conn_lost;
                     return undef;
                 }
-                substr($$bout, 0, $written, '');
             }
             if (vec($rv1, $fnoin, 1)) {
                 my $read = sysread($sftp->{ssh_in}, $$bin, 64 * 1024, length($$bin));
@@ -310,7 +337,7 @@ sub _do_io {
 			    $!);
 		    $debug & 1024 and $read and _hexdump(substr($$bin, -$read));
 		}
-                unless ($read) {
+                if (!$read and $! != Errno::EAGAIN() and $! != Errno::EINTR()) {
                     $sftp->_conn_lost;
                     return undef;
                 }
@@ -318,7 +345,7 @@ sub _do_io {
         }
         else {
             $debug and $debug & 32 and _debug "_do_io select failed: $!";
-            next if ($n < 0 and $! == Errno::EINTR());
+            next if ($n < 0 and ($! == Errno::EINTR() or $! == Errno::EAGAIN()));
             return undef;
         }
     }
